"""
Module containing supporting classes and functions used accross the project.
"""
import logging
from collections import UserDict
from functools import cached_property, lru_cache, wraps
from typing import Any, Callable, Iterable, Sequence

import numpy as np
from cachetools import LRUCache

from lymph.types import HasGetParams, HasSetParams

logger = logging.getLogger(__name__)



def check_unique_names(graph: dict):
    """Check all nodes in ``graph`` have unique names and no duplicate connections."""
    node_name_set = set()
    for (_, node_name), connections in graph.items():
        if isinstance(connections, set):
            raise TypeError("A node's connection list should not be a set (ordering)")
        if len(connections) != len(set(connections)):
            raise ValueError(f"Duplicate connections for node {node_name} in graph")
        if node_name in connections:
            raise ValueError(f"Node {node_name} is connected to itself")

        node_name_set.add(node_name)

    if len(node_name_set) != len(graph):
        raise ValueError("Node names are not unique")


def check_spsn(spsn: list[float]):
    """Check whether specificity and sensitivity are valid."""
    has_len_2 = len(spsn) == 2
    is_above_lb = np.all(np.greater_equal(spsn, 0.5))
    is_below_ub = np.all(np.less_equal(spsn, 1.))
    if not has_len_2 or not is_above_lb or not is_below_ub:
        raise ValueError(
            "For each modality provide a list of two decimals between 0.5 and 1.0 as "
            "specificity & sensitivity respectively."
        )


@lru_cache
def comp_transition_tensor(
    num_parent: int,
    num_child: int,
    is_tumor_spread: bool,
    is_growth: bool,
    spread_prob: float,
    micro_mod: float,
) -> np.ndarray:
    """Compute the transition factors of the edge.

    The returned array is of shape (p,c,c), where p is the number of states of the
    parent node and c is the number of states of the child node.

    Essentially, the tensors computed here contain most of the parametrization of
    the model. They are used to compute the transition matrix.

    This function globally computes and caches the transition tensors, such that we
    do not need to worry about deleting and recomputing them when the parameters of the
    edge change.
    """
    tensor = np.stack([np.eye(num_child)] * num_parent)

    # this should allow edges from trinary nodes to binary nodes
    pad = [0.] * (num_child - 2)

    if is_tumor_spread:
        # NOTE: Here we define how tumors spread to LNLs
        tensor[0, 0, :] = np.array([1. - spread_prob, spread_prob, *pad])
        return tensor

    if is_growth:
        # In the growth case, we can assume that two things:
        # 1. parent and child state are the same
        # 2. the child node is trinary
        tensor[1, 1, :] = np.array([0., (1 - spread_prob), spread_prob])
        return tensor

    if num_parent == 3:
        # NOTE: here we define how the micro_mod affects the spread probability
        micro_spread = spread_prob * micro_mod
        tensor[1,0,:] = np.array([1. - micro_spread, micro_spread, *pad])

        macro_spread = spread_prob
        tensor[2,0,:] = np.array([1. - macro_spread, macro_spread, *pad])

        return tensor

    tensor[1,0,:] = np.array([1. - spread_prob, spread_prob, *pad])
    return tensor


def clinical(spsn: list) -> np.ndarray:
    """Produce the confusion matrix of a clinical modality.

    A clinical modality can by definition *not* detect microscopic metastases.
    """
    check_spsn(spsn)
    sp, sn = spsn
    confusion_matrix = np.array([
        [sp     , 1. - sp],
        [sp     , 1. - sp],
        [1. - sn, sn     ],
    ])
    return confusion_matrix


def pathological(spsn: list) -> np.ndarray:
    """Produce the confusion matrix of a pathological modality.

    A pathological modality can detect microscopic disease, but is unable to
    differentiante between micro- and macroscopic involvement.
    """
    check_spsn(spsn)
    sp, sn = spsn
    confusion_matrix = np.array([
        [sp     , 1. - sp],
        [1. - sn, sn     ],
        [1. - sn, sn     ],
    ])
    return confusion_matrix


def tile_and_repeat(
    mat: np.ndarray,
    tile: tuple[int, int],
    repeat: tuple[int, int],
) -> np.ndarray:
    """Apply the numpy functions `tile`_ and `repeat`_ successively to ``mat``.

    .. _tile: https://numpy.org/doc/stable/reference/generated/numpy.tile.html
    .. _repeat: https://numpy.org/doc/stable/reference/generated/numpy.repeat.html

    >>> mat = np.array([[1, 2], [3, 4]])
    >>> tile_and_repeat(mat, (2, 2), (2, 2))
    array([[1, 1, 2, 2, 1, 1, 2, 2],
           [1, 1, 2, 2, 1, 1, 2, 2],
           [3, 3, 4, 4, 3, 3, 4, 4],
           [3, 3, 4, 4, 3, 3, 4, 4],
           [1, 1, 2, 2, 1, 1, 2, 2],
           [1, 1, 2, 2, 1, 1, 2, 2],
           [3, 3, 4, 4, 3, 3, 4, 4],
           [3, 3, 4, 4, 3, 3, 4, 4]])
    >>> tile_and_repeat(
    ...     mat=np.array([False, True], dtype=bool),
    ...     tile=(1, 2),
    ...     repeat=(1, 3),
    ... )
    array([[False, False, False,  True,  True,  True, False, False, False,
             True,  True,  True]])
    """
    tiled = np.tile(mat, tile)
    repeat_along_0 = np.repeat(tiled, repeat[0], axis=0)
    return np.repeat(repeat_along_0, repeat[1], axis=1)


@lru_cache
def get_state_idx_matrix(lnl_idx: int, num_lnls: int, num_states: int) -> np.ndarray:
    """Return the indices for the transition tensor correpsonding to ``lnl_idx``.

    >>> get_state_idx_matrix(1, 3, 2)
    array([[0, 0, 0, 0, 0, 0, 0, 0],
           [0, 0, 0, 0, 0, 0, 0, 0],
           [1, 1, 1, 1, 1, 1, 1, 1],
           [1, 1, 1, 1, 1, 1, 1, 1],
           [0, 0, 0, 0, 0, 0, 0, 0],
           [0, 0, 0, 0, 0, 0, 0, 0],
           [1, 1, 1, 1, 1, 1, 1, 1],
           [1, 1, 1, 1, 1, 1, 1, 1]])
    >>> get_state_idx_matrix(1, 2, 3)
    array([[0, 0, 0, 0, 0, 0, 0, 0, 0],
           [1, 1, 1, 1, 1, 1, 1, 1, 1],
           [2, 2, 2, 2, 2, 2, 2, 2, 2],
           [0, 0, 0, 0, 0, 0, 0, 0, 0],
           [1, 1, 1, 1, 1, 1, 1, 1, 1],
           [2, 2, 2, 2, 2, 2, 2, 2, 2],
           [0, 0, 0, 0, 0, 0, 0, 0, 0],
           [1, 1, 1, 1, 1, 1, 1, 1, 1],
           [2, 2, 2, 2, 2, 2, 2, 2, 2]])
    """
    indices = np.arange(num_states).reshape(num_states, -1)
    block = np.tile(indices, (num_states ** lnl_idx, num_states ** num_lnls))
    return np.repeat(block, num_states ** (num_lnls - lnl_idx - 1), axis=0)


def row_wise_kron(a: np.ndarray, b: np.ndarray) -> np.ndarray:
    """Compute the `kronecker product`_ of two matrices row-wise.

    .. _kronecker product: https://en.wikipedia.org/wiki/Kronecker_product

    >>> a = np.array([[1, 2], [3, 4]])
    >>> b = np.array([[5, 6], [7, 8]])
    >>> row_wise_kron(a, b)
    array([[ 5.,  6., 10., 12.],
           [21., 24., 28., 32.]])
    """
    result = np.zeros((a.shape[0], a.shape[1] * b.shape[1]))
    for i in range(a.shape[0]):
        result[i] = np.kron(a[i], b[i])

    return result


def early_late_mapping(t_stage: int | str) -> str:
    """Map the reported T-category (i.e., 1, 2, 3, 4) to "early" and "late"."""
    t_stage = int(t_stage)

    if 0 <= t_stage <= 2:
        return "early"

    if 3 <= t_stage <= 4:
        return "late"

    raise ValueError(f"Invalid T-stage: {t_stage}")


def trigger(func: callable) -> callable:
    """Decorator that runs instance's ``trigger_callbacks`` when called."""
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        result = func(self, *args, **kwargs)
        for callback in self.trigger_callbacks:
            callback()
        return result
    return wrapper


class AbstractLookupDict(UserDict):
    """Abstract ``UserDict`` subclass that can lazily and dynamically return values.

    This class is meant to be subclassed. If one wants to use the functionality
    of lazy and dynamic value retrieval, the subclass must implement a ``__missing__``
    method that returns the value for the given key and raises a ``KeyError`` if
    the value for a key cannot be computed.
    """
    trigger_callbacks: list[Callable]

    def __init__(self, dict=None, /, trigger_callbacks=None, **kwargs):
        """Use keyword arguments to set attributes of the instance.

        In contrast to the default ``UserDict`` constructor, this one instantiates
        any keyword arguments as attributes of the instance and does not put them
        into the dictionary itself.
        """
        super().__init__(dict)

        if trigger_callbacks is None:
            trigger_callbacks = []

        self.trigger_callbacks = trigger_callbacks

        for attr_name, attr_value in kwargs.items():
            if hasattr(self, attr_name):
                raise AttributeError("Cannot set attribute that already exists.")
            setattr(self, attr_name, attr_value)


    def __contains__(self, key: object) -> bool:
        """This exists to trigger ``__missing__`` when checking ``is in``."""
        if super().__contains__(key):
            return True

        if hasattr(self.__class__, "__missing__"):
            try:
                self.__class__.__missing__(self, key)
                return True
            except KeyError:
                return False

        return False


class smart_updating_dict_cached_property(cached_property):
    """Allows setting/deleting dict-like attrs by updating/clearing them."""
    def __set__(self, instance: object, value: Any) -> None:
        dict_like = self.__get__(instance)
        dict_like.clear()
        dict_like.update(value)

    def __delete__(self, instance: object) -> None:
        dict_like = self.__get__(instance)
        dict_like.clear()


def arg0_cache(maxsize: int = 128, cache_class = LRUCache) -> callable:
    """Cache a function only based on its first argument.

    One may choose which ``cache_class`` to use. This will be created with the
    argument ``maxsize``.

    Note:
        The first argument is not passed on to the decorated function. It is basically
        used as a key for the cache and it trusts the user to be sure that this is
        sufficient.
    """
    def decorator(func: callable) -> callable:
        cache = cache_class(maxsize=maxsize)

        @wraps(func)
        def wrapper(arg0, *args, **kwargs):
            if arg0 not in cache:
                cache[arg0] = func(*args, **kwargs)
            return cache[arg0]

        return wrapper

    return decorator


def dict_to_func(mapping: dict[Any, Any]) -> callable:
    """Transform a dictionary into a function.

    >>> char_map = {'a': 1, 'b': 2, 'c': 3}
    >>> char_map = dict_to_func(char_map)
    >>> char_map('a')
    1
    """
    def callable_mapping(key):
        return mapping[key]

    return callable_mapping


def popfirst(seq: Sequence[Any]) -> tuple[Any, Sequence[Any]]:
    """Return the first element of a sequence and the sequence without it.

    If the sequence is empty, the first element will be ``None`` and the second just
    the empty sequence. Example:

    >>> popfirst([1, 2, 3])
    (1, [2, 3])
    >>> popfirst([])
    (None, [])
    """
    try:
        return seq[0], seq[1:]
    except IndexError:
        return None, seq


def flatten(mapping, parent_key='', sep='_') -> dict:
    """Flatten a nested dictionary.

    >>> flatten({"a": {"b": 1, "c": 2}, "d": 3})
    {'a_b': 1, 'a_c': 2, 'd': 3}
    """
    items = []
    for k, v in mapping.items():
        new_key = f"{parent_key}{sep}{k}" if parent_key else k
        if isinstance(v, dict):
            items.extend(flatten(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


def unflatten_and_split(
    mapping: dict,
    expected_keys: list[str],
    sep: str = "_",
) -> tuple[dict, dict]:
    """Unflatten the part of a dict containing ``expected_keys`` and return the rest.

    >>> unflatten_and_split({'a_b': 1, 'a_c_x': 2, 'd_y': 3}, expected_keys=['a'])
    ({'a': {'b': 1, 'c_x': 2}}, {'d_y': 3})
    """
    split_kwargs, global_kwargs = {}, {}
    for key, value in mapping.items():
        left, _, right = key.partition(sep)
        if left not in expected_keys:
            global_kwargs[key] = value
            continue

        tmp = split_kwargs
        if left not in tmp:
            tmp[left] = {}

        tmp = tmp[left]
        tmp[right] = value

    return split_kwargs, global_kwargs


def get_params_from(
    objects: dict[str, HasGetParams],
    as_dict: bool = True,
    as_flat: bool = True,
) -> Iterable[float] | dict[str, float]:
    """Get the parameters from each ``get_params()`` method of the ``objects``."""
    params = {}
    for key, obj in objects.items():
        params[key] = obj.get_params(as_flat=as_flat)

    if as_flat or not as_dict:
        params = flatten(params)

    return params if as_dict else params.values()


def set_params_for(
    objects: dict[str, HasSetParams],
    *args: float,
    **kwargs: float,
) -> tuple[float]:
    """Pass arguments to each ``set_params()`` method of the ``objects``."""
    kwargs, global_kwargs = unflatten_and_split(kwargs, expected_keys=objects.keys())

    for key, obj in objects.items():
        obj_kwargs = global_kwargs.copy()
        obj_kwargs.update(kwargs.get(key, {}))
        args = obj.set_params(*args, **obj_kwargs)

    return args


def synchronize_params(
    get_from: dict[str, HasGetParams],
    set_to: dict[str, HasSetParams],
) -> None:
    """Get the parameters from one object and set them to another."""
    for key, obj in set_to.items():
        obj.set_params(**get_from[key].get_params(as_dict=True))


def draw_diagnoses(
    diagnose_times: list[int],
    state_evolution: np.ndarray,
    observation_matrix: np.ndarray,
    possible_diagnoses: np.ndarray,
    rng: np.random.Generator | None = None,
    seed: int = 42,
) -> np.ndarray:
    """Given the ``diagnose_times`` and a hidden ``state_evolution``, draw diagnoses."""
    if rng is None:
        rng = np.random.default_rng(seed)

    state_dists_given_time = state_evolution[diagnose_times]
    observation_dists_given_time = state_dists_given_time @ observation_matrix

    drawn_observation_idxs = [
        rng.choice(a=np.arange(len(possible_diagnoses)), p=dist)
        for dist in observation_dists_given_time
    ]
    return possible_diagnoses[drawn_observation_idxs].astype(bool)


def add_or_mult(llh: float, arr: np.ndarray, log: bool = True) -> float:
    """Add or multiply the log-likelihood with the given array."""
    if log:
        return llh + np.sum(np.log(arr))
    return llh * np.prod(arr)



"""
Methods & classes to manage matrices of the :py:class:`~lymph.models.Unilateral` class.
"""
# pylint: disable=too-few-public-methods
from __future__ import annotations

import warnings
from functools import lru_cache
from typing import Any, Iterable

import numpy as np
import pandas as pd
from cachetools import LRUCache

from lymph import graph, models
from lymph.helper import (
    AbstractLookupDict,
    arg0_cache,
    get_state_idx_matrix,
    row_wise_kron,
    tile_and_repeat,
)
from lymph.modalities import Modality


@lru_cache(maxsize=128)
def generate_transition(
    lnls: Iterable[graph.LymphNodeLevel],
    num_states: int,
) -> np.ndarray:
    """Compute the transition matrix of the lymph model."""
    lnls = list(lnls)   # necessary for `index()` call
    num_lnls = len(lnls)
    transition_matrix = np.ones(shape=(num_states**num_lnls, num_states**num_lnls))

    for i, lnl in enumerate(lnls):
        current_state_idx = get_state_idx_matrix(
            lnl_idx=i,
            num_lnls=num_lnls,
            num_states=num_states,
        )
        new_state_idx = current_state_idx.T

        # This needs to be initialized with a one where no transition happens
        # and a zero where a transition happens. This is because of how differently
        # the transition matrix entries are computed for no spread vs. spread.
        lnl_transition_matrix = new_state_idx == current_state_idx

        for edge in lnl.inc:
            if edge.is_tumor_spread:
                edge_transition_grid = edge.transition_tensor[
                    0, current_state_idx, new_state_idx
                ]
            else:
                parent_node_i = lnls.index(edge.parent)
                parent_state_idx = get_state_idx_matrix(
                    lnl_idx=parent_node_i,
                    num_lnls=num_lnls,
                    num_states=num_states,
                )
                edge_transition_grid = edge.transition_tensor[
                    parent_state_idx, current_state_idx, new_state_idx
                ]

            lnl_transition_matrix = np.where(
                # For transitions, we need to compute the probability that none of
                # the incoming edges of an LNL spread. This is done by multiplying
                # the probabilities of all edges not spreading and taking the
                # complement: 1 - (1 - p_1) * (1 - p_2) * ...
                # If an LNL remains in its current state, the probability is
                # simply the product of all incoming edges not spreading.
                new_state_idx == current_state_idx + 1,
                1 - (1 - lnl_transition_matrix) * (1 - edge_transition_grid),
                lnl_transition_matrix * edge_transition_grid,
            )

        transition_matrix *= lnl_transition_matrix

    return transition_matrix


@lru_cache(maxsize=128)
def generate_observation(
    modalities: Iterable[Modality],
    num_lnls: int,
    base: int = 2,
) -> np.ndarray:
    """Generate the observation matrix of the lymph model."""
    shape = (base ** num_lnls, 1)
    observation_matrix = np.ones(shape=shape)

    for modality in modalities:
        mod_obs_matrix = np.ones(shape=(1,1))
        for _ in range(num_lnls):
            mod_obs_matrix = np.kron(mod_obs_matrix, modality.confusion_matrix)

        observation_matrix = row_wise_kron(observation_matrix, mod_obs_matrix)

    return observation_matrix


def compute_encoding(
    lnls: list[str],
    pattern: pd.Series | dict[str, bool | int | str],
    base: int = 2,
) -> np.ndarray:
    """Compute the encoding of a particular ``pattern`` of involvement.

    A ``pattern`` holds information about the involvement of each LNL and the function
    transforms this into a binary encoding which is ``True`` for all possible complete
    states/diagnoses that are compatible with the given ``pattern``.

    In the binary case (``base=2``), the value behind ``pattern[lnl]`` can be one of
    the following things:
    - ``False``: The LNL is healthy.
    - ``"healthy"``: The LNL is healthy.
    - ``True``: The LNL is involved.
    - ``"involved"``: The LNL is involved.
    - ``pd.isna(pattern[lnl]) == True``: The involvement of the LNL is unknown.

    In the trinary case (``base=3``), the value behind ``pattern[lnl]`` can be one of
    these things:
    - ``False``: The LNL is healthy.
    - ``"healthy"``: The LNL is healthy.
    - ``True``: The LNL is involved (micro- or macroscopic).
    - ``"involved"``: The LNL is involved (micro- or macroscopic).
    - ``"micro"``: The LNL is involved microscopically only.
    - ``"macro"``: The LNL is involved macroscopically only.
    - ``"notmacro"``: The LNL is healthy or involved microscopically.

    Missing values are treated as unknown involvement.

    >>> compute_encoding(["II", "III"], {"II": True, "III": False})
    array([False, False,  True, False])
    >>> compute_encoding(["II", "III"], {"II": "involved"})
    array([False, False,  True,  True])
    >>> compute_encoding(
    ...     lnls=["II", "III"],
    ...     pattern={"II": True, "III": False},
    ...     base=3,
    ... )
    array([False, False, False,  True, False, False,  True, False, False])
    >>> compute_encoding(
    ...     lnls=["II", "III"],
    ...     pattern={"II": "micro", "III": "notmacro"},
    ...     base=3,
    ... )
    array([False, False, False,  True,  True, False, False, False, False])
    """
    num_lnls = len(lnls)
    encoding = np.ones(shape=base ** num_lnls, dtype=bool)

    if base == 2:
        element_map = {
            "healthy": np.array([True, False]),
            False: np.array([True, False]),
            "involved": np.array([False, True]),
            True: np.array([False, True]),
        }
    elif base == 3:
        element_map = {
            "healthy": np.array([True, False, False]),
            False: np.array([True, False, False]),
            "involved": np.array([False, True, True]),
            True: np.array([False, True, True]),
            "micro": np.array([False, True, False]),
            "macro": np.array([False, False, True]),
            "notmacro": np.array([True, True, False]),
        }
    else:
        raise ValueError(f"Invalid base {base}.")

    for j, lnl in enumerate(lnls):
        if lnl not in pattern or pd.isna(pattern[lnl]):
            continue

        try:
            element = element_map[pattern[lnl]]
        except KeyError as key_err:
            raise ValueError(
                f"Invalid pattern for LNL {lnl}: {pattern[lnl]}",
            ) from key_err

        encoding = np.logical_and(
            encoding,
            tile_and_repeat(
                mat=element,
                tile=(1, base ** j),
                repeat=(1, base ** (num_lnls - j - 1)),
            )[0],
        )
    return encoding


def generate_data_encoding(
    model: models.Unilateral,
    t_stage: str,
) -> np.ndarray:
    """Generate the data matrix for a specific T-stage from patient data.

    The :py:attr:`~lymph.models.Unilateral.patient_data` needs to contain the column
    ``"_model"``, which is constructed when loading the data into the model. From this,
    a data matrix is constructed for the given ``t_stage``. If ``"_BN"`` is selected,
    as T-stage, the data matrix for all patients is returned. This is mainly used for
    the computation of the Bayesian network likelihood.

    The returned matrix has the shape :math:`2^{N \\cdot \\mathcal{O}} \\times M`,
    where :math:`N` is the number of lymph node levels, :math:`\\mathcal{O}` is the
    number of diagnostic modalities and :math:`M` is the number of patients with the
    given ``t_stage`` (or just all patients).
    """
    if t_stage == "_BN":
        has_t_stage = slice(None)
    else:
        has_t_stage = model.patient_data["_model", "#", "t_stage"] == t_stage

    selected_patients = model.patient_data[has_t_stage]
    if len(selected_patients) == 0:
        raise ValueError(f"No patients with T-stage {t_stage}.")

    result = np.ones(
        shape=(model.observation_matrix().shape[1], len(selected_patients)),
        dtype=bool,
    )

    for i, (_, patient_row) in enumerate(selected_patients["_model"].iterrows()):
        patient_encoding = np.ones(shape=1, dtype=bool)
        for modality_name in model.get_all_modalities().keys():
            if modality_name not in patient_row:
                continue
            diagnose_encoding = compute_encoding(
                lnls=model.graph.lnls.keys(),
                pattern=patient_row[modality_name],
                base=2,   # observations are always binary!
            )
            patient_encoding = np.kron(patient_encoding, diagnose_encoding)

        result[:,i] = patient_encoding

    return result


class DataEncodingUserDict(AbstractLookupDict):
    """``UserDict`` that dynamically generates the data matrices for each T-stage.

    The data matrix is a binary encoding of all complete diagnoses that are plausible
    for an actual patient with possibly incomplete information. So, every columns may
    contain multiple ones, each for a complete diagnosis that could have led to the
    observed data.

    When multiplying a distribution over possible complete observations with the data
    matrix, the result is essentially an array of likelihoods for each patient.

    See Also:
        :py:attr:`~lymph.models.Unilateral.data_matrices`
    """
    model: models.Unilateral

    def __setitem__(self, __key, __value) -> None:
        warnings.warn("Setting the data matrices is not supported.")

    def __missing__(self, t_stage: str):
        """Create the data matrix for a specific T-stage if necessary."""
        try:
            self.data[t_stage] = generate_data_encoding(self.model, t_stage)
        except ValueError as val_err:
            raise KeyError(f"No data matrix for T-stage {t_stage}.") from val_err

        return self[t_stage]


def generate_diagnose(model: models.Unilateral, t_stage: str) -> np.ndarray:
    """Generate the diagnose matrix for a specific T-stage.

    The diagnose matrix is the product of the observation matrix and the data matrix
    for the given ``t_stage``.
    """
    return model.observation_matrix() @ model.data_matrices[t_stage]


cached_generate_diagnose = arg0_cache(maxsize=128, cache_class=LRUCache)(generate_diagnose)
"""Cached version of :py:func:`generate_diagnose`.

The decorated function expects an additional first argument that should be unique for
the combination of modalities and patient data. It is intended to be used with the
joint hash of the modalities (:py:meth:`.modalities_hash`) and the patient data hash
that is always precomputed when a new dataset is loaded into the model
(:py:meth:`~lymph.models.Unilateral.patient_data_hash`).
"""


class DiagnoseUserDict(AbstractLookupDict):
    """``UserDict`` that dynamically generates the diagnose matrices for each T-stage.

    A diagnose matrix for a T-stage is simply the product of the observation matrix and
    the data matrix for that T-stage. Precomputing and caching this matrix is useful,
    because it does not depend on the model parameters and can be reused until either
    the diagnostic modalities (meaning the observation matrix needs to be updated) or
    the patient data (meaning the data matrix needs to be updated) change.

    See Also:
        :py:attr:`.Unilateral.diagnose_matrices`
    """
    model: models.Unilateral

    def __setitem__(self, __key, __value) -> None:
        warnings.warn("Setting the diagnose matrices is not supported.")

    def __getitem__(self, key: Any) -> Any:
        modalities_hash = self.model.modalities_hash()
        patient_data_hash = self.model.patient_data_hash
        joint_hash = hash((modalities_hash, patient_data_hash, key))
        return cached_generate_diagnose(joint_hash, self.model, key)

    def __missing__(self, t_stage: str):
        """Create the diagnose matrix for a specific T-stage if necessary."""
        return self[t_stage]


@lru_cache
def evolve_midext(max_time: int, midext_prob: int) -> np.ndarray:
    """Compute the evolution over the state of a tumor's midline extension."""
    midext_states = np.zeros(shape=(max_time + 1, 2), dtype=float)
    midext_states[0,0] = 1.

    midext_transition_matrix = np.array([
        [1 - midext_prob, midext_prob],
        [0.             , 1.         ],
    ])

    # compute midext prob for all time steps
    for i in range(len(midext_states) - 1):
        midext_states[i+1,:] = midext_states[i,:] @ midext_transition_matrix

    return midext_states


def fast_trace(
    left: np.ndarray,
    right: np.ndarray,
) -> np.ndarray:
    """Compute the trace of a product of two matrices (``left`` and ``right``).

    This is based on the observation that the trace of a product of two matrices is
    equal to the sum of the element-wise products of the two matrices. See
    `Wikipedia <https://en.wikipedia.org/wiki/Trace_(linear_algebra)#Properties>`_ and
    `StackOverflow <https://stackoverflow.com/a/18854776>`_ for more information.
    """
    return np.sum(left.T * right, axis=0)



"""Test the graph representation class of the package."""
import unittest

import numpy as np

from lymph import graph


class ConstructBinaryGraphRepresentationTestCase(unittest.TestCase):
    """Test suite for the graph representation class."""

    def setUp(self) -> None:
        large_graph = {
            ("tumor", "T"): ["I", "II", "III", "IV", "V", "VII"],
            ("lnl", "I"): [],
            ("lnl", "II"): ["I", "III", "V"],
            ("lnl", "III"): ["IV", "V"],
            ("lnl", "IV"): [],
            ("lnl", "V"): [],
            ("lnl", "VII"): [],
        }
        medium_graph = {
            ("tumor", "T"): ["II", "III", "V"],
            ("lnl", "II"): ["III", "V"],
            ("lnl", "III"): ["V"],
            ("lnl", "V"): [],
        }
        small_graph = {
            ("tumor", "T"): ["II", "III"],
            ("lnl", "II"): ["III"],
            ("lnl", "III"): [],
        }
        self.graph_dict = large_graph
        self.graph_repr = graph.Representation(
            graph_dict=self.graph_dict,
            allowed_states=[0, 1],
            on_edge_change=[self.callback],
        )
        self.was_called = False
        self.rng = np.random.default_rng(42)

    def callback(self) -> None:
        """Callback function for the graph."""
        self.was_called = True

    def test_nodes(self) -> None:
        """Test the number of nodes."""
        self.assertEqual(len(self.graph_repr.nodes), len(self.graph_dict))
        node_names = {tpl[1] for tpl in self.graph_dict.keys()}

        for name, node in self.graph_repr.nodes.items():
            self.assertIn(name, node_names)
            if isinstance(node, graph.Tumor):
                self.assertFalse(hasattr(node, "inc"))
                self.assertTrue(hasattr(node, "out"))
            else:
                self.assertTrue(hasattr(node, "inc"))

    def test_edges(self) -> None:
        """Test the number of edges."""
        num_edges = 0
        for _, children in self.graph_dict.items():
            num_edges += len(set(children))
        self.assertEqual(len(self.graph_repr.edges), num_edges)

        for edge in self.graph_repr.edges.values():
            self.assertTrue(isinstance(edge.parent, graph.AbstractNode))
            self.assertTrue(isinstance(edge.child, graph.AbstractNode))



"""Test the trinary unilateral system."""
import unittest

import numpy as np
import pandas as pd

from lymph.graph import LymphNodeLevel
from lymph.helper import set_params_for

from . import fixtures


class TrinaryInitTestCase(fixtures.TrinaryFixtureMixin, unittest.TestCase):
    """Testing the basic initialization of a trinary model."""

    def test_is_trinary(self) -> None:
        """Test if the model is trinary."""
        self.assertTrue(self.model.is_trinary)

    def test_lnls(self):
        """Test they are all trinary lymph node levels."""
        model_allowed_states = self.model.graph.allowed_states
        self.assertEqual(len(model_allowed_states), 3)

        for lnl in self.model.graph.lnls.values():
            self.assertIsInstance(lnl, LymphNodeLevel)
            self.assertTrue(lnl.is_trinary)
            self.assertEqual(lnl.allowed_states, model_allowed_states)


class TrinaryTransitionMatrixTestCase(fixtures.TrinaryFixtureMixin, unittest.TestCase):
    """Test the transition matrix of a trinary model."""

    def setUp(self):
        super().setUp()
        params_to_set = self.create_random_params()
        self.model.set_params(**params_to_set)

    def test_edge_transition_tensors(self) -> None:
        """Test the tensors associated with each edge.

        NOTE: I am using this only in debug mode to look a the tensors. I am not sure
        how to test them yet.
        """
        base_edge_tensor = list(self.model.graph.tumor_edges.values())[0].transition_tensor
        row_sums = base_edge_tensor.sum(axis=2)
        self.assertTrue(np.allclose(row_sums, 1.0))

        lnl_edge_tensor = list(self.model.graph.lnl_edges.values())[0].transition_tensor
        row_sums = lnl_edge_tensor.sum(axis=2)
        self.assertTrue(np.allclose(row_sums, 1.0))

        growth_edge_tensor = list(self.model.graph.growth_edges.values())[0].transition_tensor
        row_sums = growth_edge_tensor.sum(axis=2)
        self.assertTrue(np.allclose(row_sums, 1.0))

    def test_transition_matrix(self) -> None:
        """Test the transition matrix of the model."""
        transition_matrix = self.model.transition_matrix()
        row_sums = transition_matrix.sum(axis=1)
        self.assertTrue(np.allclose(row_sums, 1.0))


class TrinaryObservationMatrixTestCase(fixtures.TrinaryFixtureMixin, unittest.TestCase):
    """Test the observation matrix of a trinary model."""

    def setUp(self):
        super().setUp()
        self.model.replace_all_modalities(
            self.get_modalities_subset(names=["diagnostic_consensus", "pathology"]),
        )

    def test_observation_matrix(self) -> None:
        """Test the observation matrix of the model."""
        num_lnls = len(self.model.graph.lnls)
        num = num_lnls * len(self.model.get_all_modalities())
        observation_matrix = self.model.observation_matrix()
        self.assertEqual(observation_matrix.shape, (3 ** num_lnls, 2 ** num))

        row_sums = observation_matrix.sum(axis=1)
        self.assertTrue(np.allclose(row_sums, 1.0))


class TrinaryDiagnoseMatricesTestCase(fixtures.TrinaryFixtureMixin, unittest.TestCase):
    """Test the diagnose matrix of a trinary model."""

    def setUp(self):
        super().setUp()
        self.model.replace_all_modalities(fixtures.MODALITIES)
        self.load_patient_data(filename="2021-usz-oropharynx.csv")

    def get_patient_data(self) -> pd.DataFrame:
        """Load an example dataset that has both clinical and pathology data."""
        return pd.read_csv("tests/data/2021-usz-oropharynx.csv", header=[0, 1, 2])

    def test_diagnose_matrices_shape(self) -> None:
        """Test the diagnose matrix of the model."""
        for t_stage in ["early", "late"]:
            num_lnls = len(self.model.graph.lnls)
            num_patients = (self.model.patient_data["_model", "#", "t_stage"] == t_stage).sum()
            diagnose_matrix = self.model.diagnose_matrices[t_stage]
            self.assertEqual(diagnose_matrix.shape, (3 ** num_lnls, num_patients))


class TrinaryParamAssignmentTestCase(fixtures.TrinaryFixtureMixin, unittest.TestCase):
    """Test the assignment of parameters in a trinary model."""

    def setUp(self):
        """Load patient data."""
        super().setUp()
        self.init_diag_time_dists(early="frozen", late="parametric")
        self.model.set_params(**self.create_random_params())

    def test_edge_params(self):
        """Test the assignment of edge parameters."""
        params_to_set = {}
        for edge_name, edge in self.model.graph.edges.items():
            params = edge.get_params(as_dict=True)
            for param in params:
                params_to_set[f"{edge_name}_{param}"] = self.rng.random()

        set_params_for(self.model.graph.edges, **params_to_set)
        retrieved_params = self.model.get_params(as_dict=True)
        for param in params_to_set:
            self.assertEqual(params_to_set[param], retrieved_params[param])


class TrinaryLikelihoodTestCase(fixtures.TrinaryFixtureMixin, unittest.TestCase):
    """Test the likelihood of a trinary model."""

    def setUp(self):
        """Load patient data."""
        super().setUp()
        self.model.replace_all_modalities(fixtures.MODALITIES)
        self.init_diag_time_dists(early="frozen", late="parametric")
        self.model.set_params(**self.create_random_params())
        self.load_patient_data(filename="2021-usz-oropharynx.csv")

    def test_log_likelihood_smaller_zero(self):
        """Make sure the log-likelihood is smaller than zero."""
        likelihood = self.model.likelihood(log=True, mode="HMM")
        self.assertLess(likelihood, 0.)

    def test_likelihood_invalid_params_isinf(self):
        """Make sure the likelihood is `-np.inf` for invalid parameters."""
        random_params = self.create_random_params()
        for name in random_params:
            random_params[name] += 1.
        likelihood = self.model.likelihood(
            given_params=random_params,
            log=True,
            mode="HMM",
        )
        self.assertEqual(likelihood, -np.inf)


class TrinaryRiskTestCase(fixtures.TrinaryFixtureMixin, unittest.TestCase):
    """Test the risk of a trinary model."""

    def setUp(self):
        """Load patient data."""
        super().setUp()
        self.model.replace_all_modalities(fixtures.MODALITIES)
        self.init_diag_time_dists(early="frozen", late="parametric")
        self.load_patient_data(filename="2021-usz-oropharynx.csv")

    def create_random_diagnoses(self):
        """Create a random diagnosis for each modality and LNL."""
        lnl_names = list(self.model.graph.lnls.keys())
        diagnoses = {}

        for modality in self.model.get_all_modalities():
            diagnoses[modality] = fixtures.create_random_pattern(lnl_names)

        return diagnoses

    def test_risk_is_probability(self):
        """Make sure the risk is a probability."""
        risk = self.model.risk(
            involvement=fixtures.create_random_pattern(lnls=list(self.model.graph.lnls.keys())),
            given_diagnoses=self.create_random_diagnoses(),
            given_params=self.create_random_params(),
            t_stage=self.rng.choice(["early", "late"]),
        )
        self.assertGreaterEqual(risk, 0.)
        self.assertLessEqual(risk, 1.)



"""
Test the bilateral model.
"""
import unittest

import numpy as np

from lymph import models
from lymph.helper import flatten

from . import fixtures


class BilateralInitTest(fixtures.BilateralModelMixin, unittest.TestCase):
    """Test the delegation of attrs from the unilateral class to the bilateral one."""

    def setUp(self):
        self.model_kwargs = {"is_symmetric": {
            "tumor_spread": True,
            "lnl_spread": True,
        }}
        super().setUp()
        self.load_patient_data()

    def test_delegation(self):
        """Test that the unilateral model delegates the attributes."""
        self.assertEqual(self.model.is_binary, self.model.ipsi.is_binary)
        self.assertEqual(self.model.is_trinary, self.model.ipsi.is_trinary)
        self.assertEqual(self.model.max_time, self.model.ipsi.max_time)
        self.assertEqual(list(self.model.t_stages), list(self.model.ipsi.t_stages))

    def test_transition_matrix_sync(self):
        """Make sure contra transition matrix gets recomputed when ipsi param is set."""
        ipsi_trans_mat = self.model.ipsi.transition_matrix()
        contra_trans_mat = self.model.contra.transition_matrix()
        rand_ipsi_param = self.rng.choice(list(
            self.model.ipsi.get_params(as_dict=True).keys()
        ))
        self.model.set_params(**{f"ipsi_{rand_ipsi_param}": self.rng.random()})
        self.assertFalse(np.all(
            ipsi_trans_mat == self.model.ipsi.transition_matrix()
        ))
        self.assertFalse(np.all(
            contra_trans_mat == self.model.contra.transition_matrix()
        ))

    def test_modality_sync(self):
        """Make sure the modalities are synced between the two sides."""
        self.model.set_modality("foo", spec=self.rng.uniform(), sens=self.rng.uniform())
        self.assertEqual(
            self.model.ipsi.get_modality("foo").sens,
            self.model.contra.get_modality("foo").sens,
        )
        self.assertEqual(
            self.model.ipsi.get_modality("foo").spec,
            self.model.contra.get_modality("foo").spec,
        )

    def test_asymmetric_model(self):
        """Check if different graphs work for the ipsi and contra side."""
        ipsi_graph = fixtures.get_graph("medium")
        contra_graph = fixtures.get_graph("small")

        model = models.Bilateral(
            graph_dict=ipsi_graph,
            contralateral_kwargs={"graph_dict": contra_graph},
        )

        self.assertEqual(
            list(model.ipsi.graph.nodes.keys()),
            [key[1] for key in ipsi_graph.keys()],
        )
        self.assertEqual(
            list(model.contra.graph.nodes.keys()),
            [key[1] for key in contra_graph.keys()],
        )
        self.assertEqual(
            len(model.ipsi.get_params()),
            sum(len(val) for val in ipsi_graph.values()),
        )
        self.assertEqual(
            len(model.contra.get_params()),
            sum(len(val) for val in contra_graph.values()),
        )


class ModalityDelegationTestCase(fixtures.BilateralModelMixin, unittest.TestCase):
    """Make sure the modality is delegated from the ipsi side correctly."""

    def setUp(self):
        super().setUp()
        self.model.replace_all_modalities(fixtures.MODALITIES)

    def test_modality_access(self):
        """Test that the modality can be accessed."""
        self.assertEqual(
            self.model.get_modality("CT").sens,
            self.model.ipsi.get_modality("CT").sens,
        )
        self.assertEqual(
            self.model.get_modality("FNA").spec,
            self.model.ipsi.get_modality("FNA").spec,
        )

    def test_modality_delete(self):
        """Test that the modality can be deleted."""
        self.model.del_modality("CT")
        self.assertNotIn("CT", self.model.get_all_modalities())
        self.assertNotIn("CT", self.model.ipsi.get_all_modalities())
        self.assertNotIn("CT", self.model.contra.get_all_modalities())

    def test_modality_update(self):
        """Test that the modality can be updated."""
        old_mod = self.model.get_modality("CT")
        self.model.set_modality("CT", spec=old_mod.spec, sens=0.8)
        self.assertEqual(
            self.model.get_modality("CT").sens,
            self.model.ipsi.get_modality("CT").sens,
        )
        self.assertEqual(
            self.model.get_modality("CT").sens,
            self.model.contra.get_modality("CT").sens,
        )

    def test_modality_reset(self):
        """Test resetting the modalities also works."""
        self.model.set_modality("foo", spec=0.8, sens=0.9)
        self.assertEqual(
            self.model.get_modality("foo").sens,
            self.model.ipsi.get_modality("foo").sens,
        )
        self.assertEqual(
            self.model.get_modality("foo").spec,
            self.model.contra.get_modality("foo").spec,
        )

    def test_diag_time_dists_delegation(self):
        """Test that the diagnose time distributions are delegated."""
        self.assertEqual(
            list(self.model.get_distribution("early").pmf),
            list(self.model.ipsi.get_distribution("early").pmf),
        )
        self.assertEqual(
            list(self.model.get_distribution("late").get_params()),
            list(self.model.ipsi.get_distribution("late").get_params()),
        )
        self.assertEqual(
            list(self.model.get_distribution("early").pmf),
            list(self.model.contra.get_distribution("early").pmf),
        )
        self.assertEqual(
            list(self.model.get_distribution("late").get_params()),
            list(self.model.contra.get_distribution("late").get_params()),
        )


class NoSymmetryParamsTestCase(fixtures.BilateralModelMixin, unittest.TestCase):
    """Test the parameter assignment when the model is not symmetric"""

    def setUp(self):
        self.model_kwargs = {
            "is_symmetric": {
                "tumor_spread": False,
                "lnl_spread": False,
            }
        }
        super().setUp()

    def test_get_params_as_args(self):
        """Test that the parameters can be retrieved."""
        ipsi_args = list(self.model.ipsi.get_params(as_dict=False))
        contra_args = list(self.model.contra.get_params(as_dict=False))
        both_args = list(self.model.get_params(as_dict=False))
        num_param_dists = len(self.model.get_distribution_params())
        # need plus one, because distribution's parameter is accounted for twice
        self.assertEqual(len(ipsi_args) + len(contra_args), len(both_args) + 1)
        self.assertEqual(
            [*ipsi_args[:-num_param_dists], *contra_args[:-num_param_dists]],
            both_args[:-num_param_dists],
        )

    def test_get_params_as_dict(self):
        """Test that the parameters can be retrieved."""
        ipsi_dict = self.model.ipsi.get_params(as_dict=True)
        contra_dict = self.model.contra.get_params(as_dict=True)
        both_dict = self.model.get_params(as_dict=True, as_flat=False)
        dist_param_keys = self.model.get_distribution_params().keys()

        for key in dist_param_keys:
            ipsi_dict.pop(key)
            contra_dict.pop(key)

        self.assertEqual(ipsi_dict, flatten(both_dict["ipsi"]))
        self.assertEqual(contra_dict, flatten(both_dict["contra"]))

    def test_set_params_as_args(self):
        """Test that the parameters can be set."""
        ipsi_tumor_spread_args = self.rng.uniform(size=len(self.model.ipsi.graph.tumor_edges))
        ipsi_lnl_spread_args = self.rng.uniform(size=len(self.model.ipsi.graph.lnl_edges))
        contra_tumor_spread_args = self.rng.uniform(size=len(self.model.contra.graph.tumor_edges))
        contra_lnl_spread_args = self.rng.uniform(size=len(self.model.contra.graph.lnl_edges))
        dist_params = self.rng.uniform(size=len(self.model.get_distribution_params()))

        self.model.set_params(
            *ipsi_tumor_spread_args,
            *contra_tumor_spread_args,
            *ipsi_lnl_spread_args,
            *contra_lnl_spread_args,
            *dist_params,
        )
        self.assertEqual(
            [*ipsi_tumor_spread_args, *ipsi_lnl_spread_args, *dist_params],
            list(self.model.ipsi.get_params(as_dict=False)),
        )
        self.assertEqual(
            [*contra_tumor_spread_args, *contra_lnl_spread_args, *dist_params],
            list(self.model.contra.get_params(as_dict=False)),
        )
        self.assertEqual(
            list(self.model.ipsi.get_distribution("late").get_params())[0],
            list(self.model.contra.get_distribution("late").get_params())[0],
        )

    def test_set_params_as_dict(self):
        """Test that the parameters can be set via keyword arguments."""
        params_to_set = {k: self.rng.uniform() for k in self.model.get_params().keys()}
        self.model.set_params(**params_to_set)
        self.assertEqual(params_to_set, self.model.get_params())


class SymmetryParamsTestCase(fixtures.BilateralModelMixin, unittest.TestCase):
    """Test the parameter assignment when the model is symmetric."""

    def setUp(self):
        self.model_kwargs = {
            "is_symmetric": {
                "tumor_spread": True,
                "lnl_spread": True,
            }
        }
        super().setUp()

    def test_get_params_as_args(self):
        """Test that the parameters can be retrieved."""
        ipsi_args = list(self.model.ipsi.get_params(as_dict=False))
        contra_args = list(self.model.contra.get_params(as_dict=False))
        both_args = list(self.model.get_params(as_dict=False))
        self.assertEqual(ipsi_args, both_args)
        self.assertEqual(contra_args, both_args)

    def test_get_params_as_dict(self):
        """Test that the parameters can be retrieved."""
        ipsi_dict = self.model.ipsi.get_params()
        contra_dict = self.model.contra.get_params()
        both_dict = self.model.get_params()
        self.assertEqual(ipsi_dict, both_dict)
        self.assertEqual(contra_dict, both_dict)

    def test_set_params_as_args(self):
        """Test that the parameters can be set."""
        args_to_set = [self.rng.uniform() for _ in self.model.ipsi.get_params(as_dict=False)]
        self.model.set_params(*args_to_set)
        self.assertEqual(args_to_set, list(self.model.contra.get_params().values()))

    def test_set_params_as_dict(self):
        """Test that the parameters can be set via keyword arguments."""
        params_to_set = {k: self.rng.uniform() for k in self.model.contra.get_params()}
        self.model.set_params(**params_to_set)
        self.assertEqual(params_to_set, self.model.ipsi.get_params())


class LikelihoodTestCase(fixtures.BilateralModelMixin, unittest.TestCase):
    """Check that the (log-)likelihood is computed correctly."""

    def setUp(self):
        super().setUp()
        self.model.replace_all_modalities(fixtures.MODALITIES)
        self.load_patient_data()

    def test_compute_likelihood_twice(self):
        """Test that the likelihood is computed correctly."""
        first_llh = self.model.likelihood(log=True)
        second_llh = self.model.likelihood(log=True)
        self.assertEqual(first_llh, second_llh)


class RiskTestCase(fixtures.BilateralModelMixin, unittest.TestCase):
    """Check that the risk is computed correctly."""

    def setUp(self):
        super().setUp()
        self.model.replace_all_modalities(fixtures.MODALITIES)

    def create_random_diagnoses(self):
        """Create a random diagnosis for each modality and LNL."""
        diagnoses = {}

        for side in ["ipsi", "contra"]:
            diagnoses[side] = {}
            side_model = getattr(self.model, side)
            lnl_names = side_model.graph.lnls.keys()
            for modality in side_model.get_all_modalities():
                diagnoses[side][modality] = fixtures.create_random_pattern(lnl_names)

        return diagnoses

    def test_posterior_state_dist(self):
        """Test that the posterior state distribution is computed correctly."""
        num_states = len(self.model.ipsi.graph.state_list)
        random_parameters = self.create_random_params()
        random_diagnoses = self.create_random_diagnoses()

        posterior = self.model.comp_posterior_joint_state_dist(
            given_params=random_parameters,
            given_diagnoses=random_diagnoses,
        )
        self.assertEqual(posterior.shape, (num_states, num_states))
        self.assertEqual(posterior.dtype, float)
        self.assertTrue(np.isclose(posterior.sum(), 1.))

    def test_risk(self):
        """Test that the risk is computed correctly."""
        random_parameters = self.create_random_params()
        random_diagnoses = self.create_random_diagnoses()
        random_pattern = {
            "ipsi": fixtures.create_random_pattern(self.model.ipsi.graph.lnls.keys()),
            "contra": fixtures.create_random_pattern(self.model.contra.graph.lnls.keys()),
        }
        random_t_stage = self.rng.choice(["early", "late"])

        risk = self.model.risk(
            involvement=random_pattern,
            given_params=random_parameters,
            given_diagnoses=random_diagnoses,
            t_stage=random_t_stage,
        )
        self.assertLessEqual(risk, 1.)
        self.assertGreaterEqual(risk, 0.)


class DataGenerationTestCase(fixtures.BilateralModelMixin, unittest.TestCase):
    """Check the binary model's data generation method."""

    def setUp(self):
        super().setUp()
        self.model.replace_all_modalities(fixtures.MODALITIES)
        self.init_diag_time_dists(early="frozen", late="parametric")
        self.model.set_params(**self.create_random_params())

    def test_generate_data(self):
        """Check bilateral data generation."""
        dataset = self.model.draw_patients(
            num=10000,
            stage_dist=[0.5, 0.5],
            rng=self.rng,
        )

        for mod in self.model.get_all_modalities():
            self.assertIn(mod, dataset)
            for side in ["ipsi", "contra"]:
                self.assertIn(side, dataset[mod])
                for lnl in self.model.ipsi.graph.lnls.keys():
                    self.assertIn(lnl, dataset[mod][side])

        self.assertAlmostEqual(
            (dataset["tumor", "1", "t_stage"] == "early").mean(), 0.5,
            delta=0.02
        )



"""
Module implementing management of the diagnostic modalities.

This allows the user to define diagnostic modalities and their sensitivity/specificity
values. This is necessary to compute the likelihood of a dataset (that was created by
recoding the output of diagnostic modalities), given the model and its parameters
(which we want to learn).
"""
from __future__ import annotations

import warnings
from abc import ABC, abstractmethod
from typing import Literal, TypeVar

import numpy as np


class Modality:
    """Stores the confusion matrix of a diagnostic modality."""
    def __init__(
        self,
        spec: float,
        sens: float,
        is_trinary: bool = False,
    ) -> None:
        if not (0. <= sens <= 1. and 0. <= spec <= 1.):
            raise ValueError("Senstivity and specificity must be between 0 and 1.")

        self.spec = spec
        self.sens = sens
        self.is_trinary = is_trinary


    def __hash__(self) -> int:
        """Return a hash of the modality.

        This is computed from the confusion matrix of the modality.
        """
        return hash(self.confusion_matrix.tobytes())


    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Modality):
            return False

        return np.all(self.confusion_matrix == other.confusion_matrix)


    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}("
            f"spec={self.spec!r}, "
            f"sens={self.sens!r}, "
            f"is_trinary={self.is_trinary!r})"
        )


    @property
    def spec(self) -> float:
        """Return the specificity of the modality."""
        return self._spec

    @spec.setter
    def spec(self, value: float) -> None:
        """Set the specificity of the modality."""
        if not 0. <= value <= 1.:
            raise ValueError("Specificity must be between 0 and 1.")

        if hasattr(self, "_confusion_matrix"):
            del self._confusion_matrix

        self._spec = value


    @property
    def sens(self) -> float:
        """Return the sensitivity of the modality."""
        return self._sens

    @sens.setter
    def sens(self, value: float) -> None:
        """Set the sensitivity of the modality."""
        if not 0. <= value <= 1.:
            raise ValueError("Sensitivity must be between 0 and 1.")

        if hasattr(self, "_confusion_matrix"):
            del self._confusion_matrix

        self._sens = value


    def compute_confusion_matrix(self) -> np.ndarray:
        """Compute the confusion matrix of the modality."""
        return np.array([
            [self.spec, 1. - self.spec],
            [1. - self.sens, self.sens],
        ])

    @property
    def confusion_matrix(self) -> np.ndarray:
        """Return the confusion matrix of the modality."""
        if not hasattr(self, '_confusion_matrix'):
            self.confusion_matrix = self.compute_confusion_matrix()

        if self.is_trinary and not self._confusion_matrix.shape[0] == 3:
            self.confusion_matrix = self.compute_confusion_matrix()

        return self._confusion_matrix

    @confusion_matrix.setter
    def confusion_matrix(self, value: np.ndarray) -> None:
        """Set the confusion matrix of the modality."""
        self.check_confusion_matrix(value)
        self._confusion_matrix = value

    def check_confusion_matrix(self, value: np.ndarray) -> None:
        """Check if the confusion matrix is valid."""
        row_sums = np.sum(value, axis=1)
        if not np.allclose(row_sums, 1.):
            raise ValueError("Rows of confusion matrix must sum to one.")

        if not np.all(np.greater_equal(value, 0.)):
            raise ValueError("Confusion matrix must be non-negative.")

        if not np.all(np.less_equal(value, 1.)):
            raise ValueError("Confusion matrix must be less than or equal to one.")

        if self.is_trinary and value.shape[0] != 3:
            raise ValueError("Confusion matrix must have 3 rows for trinary models.")

        if not self.is_trinary and value.shape[0] != 2:
            raise ValueError("Confusion matrix must have 2 rows for binary models.")


class Clinical(Modality):
    """Stores the confusion matrix of a clinical modality."""
    def compute_confusion_matrix(self) -> np.ndarray:
        """Compute the confusion matrix of the clinical modality."""
        binary_confusion_matrix = super().compute_confusion_matrix()
        if not self.is_trinary:
            return binary_confusion_matrix

        return np.vstack([binary_confusion_matrix[0], binary_confusion_matrix])


class Pathological(Modality):
    """Stores the confusion matrix of a pathological modality."""
    def compute_confusion_matrix(self) -> np.ndarray:
        """Return the confusion matrix of the pathological modality."""
        binary_confusion_matrix = super().compute_confusion_matrix()
        if not self.is_trinary:
            return binary_confusion_matrix

        return np.vstack([binary_confusion_matrix, binary_confusion_matrix[1]])



MC = TypeVar("MC", bound="Composite")

class Composite(ABC):
    """Abstract base class implementing the composite pattern for diagnostic modalities.

    Any class inheriting from this class should be able to handle the definition of
    diagnostic modalities and their sensitivity/specificity values,
    """
    _is_trinary: bool
    _modalities: dict[str, Modality]    # only for leaf nodes
    _modality_children: dict[str, Composite]

    def __init__(
        self: MC,
        modality_children: dict[str, Composite] = None,
        is_modality_leaf: bool = False,
    ) -> None:
        """Initialize the modality composite."""
        if modality_children is None:
            modality_children = {}

        if is_modality_leaf:
            self._modalities = {}
            modality_children = {}   # ignore any provided children

        self._modality_children = modality_children
        super().__init__()


    @property
    def _is_modality_leaf(self: MC) -> bool:
        """Return whether the composite is a leaf node."""
        if len(self._modality_children) > 0:
            return False

        if not hasattr(self, "_modalities"):
            raise AttributeError(f"{self} has no children and no modalities.")

        return True


    @property
    @abstractmethod
    def is_trinary(self: MC) -> bool:
        """Return whether the modality is trinary."""


    def get_modality(self: MC, name: str) -> Modality:
        """Return the modality with the given ``name``."""
        return self.get_all_modalities()[name]


    def get_all_modalities(self: MC) -> dict[str, Modality]:
        """Return all modalities of the composite.

        This will issue a warning if it finds that not all modalities of the composite
        are equal. Note that it will always return the modalities of the first child.
        This means one should NOT try to set the modalities via the returned dictionary
        of this method. Instead, use the :py:meth:`.set_modality` method.
        """
        if self._is_modality_leaf:
            return self._modalities

        child_keys = list(self._modality_children.keys())
        first_child = self._modality_children[child_keys[0]]
        firs_modalities = first_child.get_all_modalities()
        are_all_equal = True
        for key in child_keys[1:]:
            other_child = self._modality_children[key]
            are_all_equal &= firs_modalities == other_child.get_all_modalities()

        if not are_all_equal:
            warnings.warn("Not all modalities are equal. Returning first one.")

        return firs_modalities


    def set_modality(
        self,
        name: str,
        spec: float,
        sens: float,
        kind: Literal["clinical", "pathological"] = "clinical",
    ) -> None:
        """Set the modality with the given ``name``."""
        if self._is_modality_leaf:
            cls = Pathological if kind == "pathological" else Clinical
            self._modalities[name] = cls(spec, sens, self.is_trinary)

        else:
            for child in self._modality_children.values():
                child.set_modality(name, spec, sens, kind)


    def del_modality(self: MC, name: str) -> None:
        """Delete the modality with the given ``name``."""
        if self._is_modality_leaf:
            del self._modalities[name]

        else:
            for child in self._modality_children.values():
                child.del_modality(name)


    def replace_all_modalities(self: MC, modalities: dict[str, Modality]) -> None:
        """Replace all modalities of the composite with new ``modalities``."""
        if self._is_modality_leaf:
            self.clear_modalities()
            for name, modality in modalities.items():
                kind = "pathological" if isinstance(modality, Pathological) else "clinical"
                self.set_modality(name, modality.spec, modality.sens, kind)

        else:
            for child in self._modality_children.values():
                child.replace_all_modalities(modalities)


    def modalities_hash(self: MC) -> int:
        """Compute a hash from all stored modalities.

        See the :py:meth:`.Modality.__hash__` method for more information.
        """
        hash_res = 0
        if self._is_modality_leaf:
            for name, modality in self._modalities.items():
                hash_res = hash((hash_res, name, hash(modality)))

        else:
            for child in self._modality_children.values():
                hash_res = hash((hash_res, child.modalities_hash()))

        return hash_res


    def clear_modalities(self: MC) -> None:
        """Clear all modalities of the composite."""
        if self._is_modality_leaf:
            self._modalities.clear()

        else:
            for child in self._modality_children.values():
                child.clear_modalities()



"""
Make sure the models work with the emcee package.
"""
import unittest

import emcee
import numpy as np

from . import fixtures


class UnilateralEmceeTestCase(fixtures.BinaryUnilateralModelMixin, unittest.TestCase):
    """Test the emcee package with the Unilateral model."""

    def setUp(self):
        super().setUp(graph_size="small")
        self.model.set_modality("PET", spec=0.86, sens=0.79)
        self.load_patient_data(filename="2021-usz-oropharynx.csv")


    def test_emcee(self):
        """Test the emcee package with the Unilateral model."""
        nwalkers, ndim = 50, len(self.model.get_params())
        nsteps = 100
        initial = self.rng.uniform(size=(nwalkers, ndim))

        sampler = emcee.EnsembleSampler(
            nwalkers=nwalkers,
            ndim=ndim,
            log_prob_fn=self.model.likelihood,
            moves=[(emcee.moves.DEMove(), 0.8), (emcee.moves.DESnookerMove(), 0.2)],
            parameter_names=list(self.model.get_params().keys()),
        )
        sampler.run_mcmc(initial, nsteps, progress=True)
        samples = sampler.get_chain(discard=int(0.9*nsteps), flat=True)
        self.assertGreater(sampler.acceptance_fraction.mean(), 0.2)
        self.assertLess(sampler.acceptance_fraction.mean(), 0.5)
        self.assertTrue(np.all(samples.mean(axis=0) >= 0.0))
        self.assertTrue(np.all(samples.mean(axis=0) <= 1.0))



"""
Module defining the nodes and edges of the graph representing the lymphatic system.

Anything related to the network of nodes and edges is defined here. This includes the
nodes themselves (either :py:class:`~Tumor` or :py:class:`~LymphNodeLevel`), the edges
(:py:class:`~Edge`), and the graph (:py:class:`~Representation`).

The nodes and edges are used to define the structure of the graph, which may then be
accessed via the :py:class:`~Representation` class. This in turn is then used to
compute e.g. the transition matrix of the model.
"""
from __future__ import annotations

import base64
import warnings
from itertools import product
from typing import Iterable

import numpy as np

from lymph.helper import (
    check_unique_names,
    comp_transition_tensor,
    flatten,
    popfirst,
    set_params_for,
    trigger,
)


class AbstractNode:
    """Abstract base class for nodes in the graph reprsenting the lymphatic system."""
    def __init__(
        self,
        name: str,
        state: int,
        allowed_states: list[int] | None = None,
    ) -> None:
        """Make a new node.

        Upon initialization, the ``name`` and ``state`` of the node must be provided.
        The ``state`` must be one of the ``allowed_states``. The constructor makes sure that
        the ``allowed_states`` are a list of ints, even when, e.g., a tuple of floats
        is provided.
        """
        self.name = name

        if allowed_states is None:
            allowed_states = [0, 1]

        _allowed_states = []
        for s in allowed_states:
            try:
                _allowed_states.append(int(s))
            except ValueError as val_err:
                raise ValueError("Allowed states must be castable to int") from val_err

        self.allowed_states = _allowed_states
        self.state = state

        # nodes can have outgoing edge connections
        self.out: list[Edge] = []


    def __str__(self) -> str:
        """Return a string representation of the node."""
        return self.name

    def __repr__(self) -> str:
        """Return a string representation of the node."""
        cls_name = type(self).__name__
        return (
            f"{cls_name}("
            f"name={self.name!r}, "
            f"state={self.state!r}, "
            f"allowed_states={self.allowed_states!r})"
        )

    def __hash__(self) -> int:
        """Return a hash of the node's name and state."""
        return hash((self.name, self.state, tuple(self.allowed_states)))


    @property
    def name(self) -> str:
        """Return the name of the node."""
        return self._name

    @name.setter
    def name(self, new_name: str) -> None:
        """Set the name of the node."""
        self._name = str(new_name)


    @property
    def state(self) -> int:
        """Return the state of the node."""
        return self._state

    @state.setter
    def state(self, new_state: int) -> None:
        """Set the state of the node."""
        new_state = int(new_state)

        if new_state not in self.allowed_states:
            raise ValueError("State of node must be one of the allowed states")

        self._state = new_state


    def comp_obs_prob(
        self,
        obs: int,
        obs_table: np.ndarray,
        log: bool = False,
    ) -> float:
        """Compute the probability of the diagnosis ``obs``, given the current state.

        The ``obs_table`` is a 2D array with the rows corresponding to the states and
        the columns corresponding to the observations. It encodes for each state and
        diagnosis the corresponding probability.
        """
        if obs is None or np.isnan(obs):
            return 0 if log else 1.
        obs_prob = obs_table[self.state, int(obs)]
        return np.log(obs_prob) if log else obs_prob


class Tumor(AbstractNode):
    """A tumor in the graph representation of the lymphatic system."""
    def __init__(self, name: str, state: int = 1) -> None:
        """Create a new tumor node.

        It can only ever be in one ``state``, which is implemented such that the
        ``allowed_states`` are set to ``[state]``.
        """
        allowed_states = [state]
        super().__init__(name, state, allowed_states)


    def __str__(self):
        """Print basic info"""
        return f"Tumor '{super().__str__()}'"


class LymphNodeLevel(AbstractNode):
    """A lymph node level (LNL) in the graph representation of the lymphatic system."""
    def __init__(
        self,
        name: str,
        state: int = 0,
        allowed_states: list[int] | None = None,
    ) -> None:
        """Create a new lymph node level."""

        super().__init__(name, state, allowed_states)

        # LNLs can also have incoming edge connections
        self.inc: list[Edge] = []


    @classmethod
    def binary(cls, name: str, state: int = 0) -> LymphNodeLevel:
        """Create a new binary LNL."""
        return cls(name, state, [0, 1])

    @classmethod
    def trinary(cls, name: str, state: int = 0) -> LymphNodeLevel:
        """Create a new trinary LNL."""
        return cls(name, state, [0, 1, 2])


    def __str__(self):
        """Print basic info"""
        narity = "binary" if self.is_binary else "trinary"
        return f"{narity} LNL '{super().__str__()}'"


    @property
    def is_binary(self) -> bool:
        """Return whether the node is binary."""
        return len(self.allowed_states) == 2


    @property
    def is_trinary(self) -> bool:
        """Return whether the node is trinary."""
        return len(self.allowed_states) == 3


    def comp_bayes_net_prob(self, log: bool = False) -> float:
        """Compute the Bayesian network's probability for the current state."""
        if self.is_trinary:
            raise NotImplementedError("Trinary nodes are not yet supported")

        res = (-1) ** self.state
        for edge in self.inc:
            parent_state = 0 if isinstance(edge.parent, Tumor) else edge.parent.state
            res *= edge.transition_tensor[parent_state, 0, 0]

        res += self.state
        return np.log(res) if log else res


    def comp_trans_prob(self, new_state: int) -> float:
        """Compute the hidden Markov model's transition probability to a ``new_state``."""
        if new_state == self.state:
            stay_prob = 1.
            for edge in self.inc:
                edge_prob = edge.transition_tensor[edge.parent.state, self.state, new_state]
                stay_prob *= edge_prob
            return stay_prob

        transition_prob = 0.
        for edge in self.inc:
            edge_prob = edge.transition_tensor[edge.parent.state, self.state, new_state]
            transition_prob = 1. - (1. - transition_prob) * (1. - edge_prob)


class Edge:
    """This class represents an arc in the graph representation of the lymph system."""
    def __init__(
        self,
        parent: Tumor | LymphNodeLevel,
        child: LymphNodeLevel,
        spread_prob: float = 0.,
        micro_mod: float = 1.,
        callbacks: list[callable] | None = None,
    ) -> None:
        """Create a new edge between two nodes.

        The ``parent`` node must be a :py:class:`Tumor` or a :py:class:`LymphNodeLevel`,
        and the ``child`` node must be a :py:class:`LymphNodeLevel`.

        The ``spread_prob`` parameter is the probability of a tumor or involved LNL to
        spread to the next LNL. The ``micro_mod`` parameter is a modifier for the spread
        probability in case of only a microscopic node involvement.
        """
        self.trigger_callbacks = []
        if callbacks is not None:
            self.trigger_callbacks += callbacks

        self.parent: Tumor | LymphNodeLevel = parent
        self.child: LymphNodeLevel = child

        if (
            not isinstance(self.parent, Tumor)
            and self.parent.is_trinary
            and not self.is_growth
        ):
            self.micro_mod = micro_mod

        self.spread_prob = spread_prob


    def __str__(self) -> str:
        """Print basic info."""
        return f"Edge {self.get_name(middle=' to ')}"

    def __repr__(self) -> str:
        """Print basic info."""
        cls_name = type(self).__name__
        return (
            f"{cls_name}("
            f"parent={self.parent!r}, "
            f"child={self.child!r}, "
            f"spread_prob={self.spread_prob!r}, "
            f"micro_mod={self.micro_mod!r})"
        )

    def __hash__(self) -> int:
        """Return a hash of the edge's transition tensor."""
        return hash((self.get_name(), self.transition_tensor.tobytes()))


    @property
    def parent(self) -> Tumor | LymphNodeLevel:
        """Return the parent node that drains lymphatically via the edge."""
        return self._parent

    @parent.setter
    def parent(self, new_parent: Tumor | LymphNodeLevel) -> None:
        """Set the parent node of the edge."""
        if hasattr(self, '_parent'):
            self.parent.out.remove(self)

        if not issubclass(new_parent.__class__, AbstractNode):
            raise TypeError("Start must be instance of Node!")

        self._parent = new_parent
        self.parent.out.append(self)


    @property
    def child(self) -> LymphNodeLevel:
        """Return the child node of the edge, receiving lymphatic drainage."""
        return self._child

    @child.setter
    def child(self, new_child: LymphNodeLevel) -> None:
        """Set the end (child) node of the edge."""
        if hasattr(self, '_child'):
            self.child.inc.remove(self)

        if not isinstance(new_child, LymphNodeLevel):
            raise TypeError("End must be instance of Node!")

        self._child = new_child
        self.child.inc.append(self)


    def get_name(self, middle='to') -> str:
        """Return the name of the edge.

        An edge's name is simply the name of the parent node and the child node,
        connected by the string provided via the ``middle`` argument.

        This is used to identify and assign spread probabilities to it e.g. in the
        :py:class:`~models.Unilateral.set_params()` method and elsewhere.

        >>> lnl_II = LymphNodeLevel("II")
        >>> lnl_III = LymphNodeLevel("III")
        >>> edge = Edge(lnl_II, lnl_III)
        >>> edge.get_name()
        'IItoIII'
        >>> edge.get_name(middle='->')
        'II->III'
        """
        if self.is_growth:
            return self.parent.name

        return f"{self.parent.name}{middle}{self.child.name}"


    @property
    def is_growth(self) -> bool:
        """Check if this edge represents a node's growth."""
        return self.parent == self.child


    @property
    def is_tumor_spread(self) -> bool:
        """Check if this edge represents spread from a tumor to an LNL."""
        return isinstance(self.parent, Tumor)


    def get_micro_mod(self) -> float:
        """Return the spread probability."""
        if (
            not hasattr(self, "_micro_mod")
            or isinstance(self.parent, Tumor)
            or self.parent.is_binary
        ):
            self._micro_mod = 1.
        return self._micro_mod

    @trigger
    def set_micro_mod(self, new_micro_mod: float | None) -> None:
        """Set the spread modifier for LNLs with microscopic involvement."""
        if new_micro_mod is None:
            return

        if isinstance(self.parent, Tumor) or self.parent.is_binary:
            warnings.warn("Microscopic spread modifier is not used for binary nodes!")

        if not 0. <= new_micro_mod <= 1.:
            raise ValueError("Microscopic spread modifier must be between 0 and 1!")

        self._micro_mod = new_micro_mod

    micro_mod = property(
        fget=get_micro_mod,
        fset=set_micro_mod,
        doc="Parameter modifying spread probability in case of macroscopic involvement",
    )


    def get_spread_prob(self) -> float:
        """Return the spread probability."""
        if not hasattr(self, "_spread_prob"):
            self._spread_prob = 0.
        return self._spread_prob

    @trigger
    def set_spread_prob(self, new_spread_prob: float | None) -> None:
        """Set the spread probability of the edge."""
        if new_spread_prob is None:
            return

        if not 0. <= new_spread_prob <= 1.:
            raise ValueError("Spread probability must be between 0 and 1!")

        self._spread_prob = new_spread_prob

    spread_prob = property(
        fget=get_spread_prob,
        fset=set_spread_prob,
        doc="Spread probability of the edge",
    )


    def get_params(
        self,
        as_dict: bool = True,
        **_kwargs,
    ) -> Iterable[float] | dict[str, float]:
        """Return the value of the parameter ``param`` or all params in a dict.

        See Also:
            :py:meth:`lymph.diagnose_times.Distribution.get_params`
            :py:meth:`lymph.diagnose_times.DistributionsUserDict.get_params`
            :py:meth:`lymph.models.Unilateral.get_params`
            :py:meth:`lymph.models.Bilateral.get_params`
        """
        if self.is_growth:
            params = {"growth": self.get_spread_prob()}
            return params if as_dict else params.values()

        params = {"spread": self.get_spread_prob()}
        if self.child.is_trinary and not self.is_tumor_spread:
            params["micro"] = self.get_micro_mod()

        return params if as_dict else params.values()


    def set_params(self, *args, **kwargs) -> tuple[float]:
        """Set the values of the edge's parameters.

        If provided as positional arguments, the edge connects to a trinary node, and
        is not a growth node, the first argument is the spread probability and the
        second argument is the microscopic spread modifier. Otherwise it only consumes
        one argument, which is the growth or spread probability.

        Keyword arguments (i.e., ``"growth"``, ``"spread"``, and ``"micro"``) override
        positional arguments. Unused args are returned.

        >>> edge = Edge(LymphNodeLevel("II", allowed_states=[0, 1, 2]), LymphNodeLevel("III"))
        >>> _ = edge.set_params(0.1, 0.2)
        >>> edge.spread_prob
        0.1
        >>> edge.micro_mod
        0.2
        >>> _ = edge.set_params(spread=0.3, micro=0.4)
        >>> edge.spread_prob
        0.3
        >>> edge.micro_mod
        0.4
        """
        first, args = popfirst(args)
        value = first or self.get_spread_prob()

        if self.is_growth:
            self.set_spread_prob(kwargs.get("growth", value))
        else:
            self.set_spread_prob(kwargs.get("spread", value))

        if (
            not isinstance(self.parent, Tumor)
            and self.parent.is_trinary
            and not self.is_growth
        ):
            first, args = popfirst(args)
            value = first or self.get_micro_mod()
            self.set_micro_mod(kwargs.get("micro", value))

        return args


    @property
    def transition_tensor(self) -> np.ndarray:
        """Return the transition tensor of the edge.

        See Also:
            :py:func:`lymph.helper.comp_transition_tensor`
        """
        return comp_transition_tensor(
            num_parent=len(self.parent.allowed_states),
            num_child=len(self.child.allowed_states),
            is_tumor_spread=self.is_tumor_spread,
            is_growth=self.is_growth,
            spread_prob=self.spread_prob,
            micro_mod=self.micro_mod,
        )


class Representation:
    """Class holding the graph structure of the model.

    This class allows accessing the connected nodes (:py:class:`Tumor` and
    :py:class:`LymphNodeLevel`) and edges (:py:class:`Edge`) of the :py:mod:`models`.
    """
    def __init__(
        self,
        graph_dict: dict[tuple[str], list[str]],
        tumor_state: int | None = None,
        allowed_states: list[int] | None = None,
        on_edge_change: list[callable] | None = None,
    ) -> None:
        """Create a new graph representation of nodes and edges.

        The ``graph_dict`` is a dictionary that defines which nodes are created and
        with what edges they are connected. The keys of the dictionary are tuples of
        the form ``(node_type, node_name)``. The ``node_type`` can be either ``"tumor"``
        or ``"lnl"``. The ``node_name`` is a string that uniquely identifies the node.
        The values of the dictionary are lists of node names to which the key node
        should be connected.
        """
        if allowed_states is None:
            allowed_states = [0, 1]

        if tumor_state is None:
            tumor_state = allowed_states[-1]

        check_unique_names(graph_dict)
        self._init_nodes(graph_dict, tumor_state, allowed_states)
        self._init_edges(graph_dict, on_edge_change)


    def _init_nodes(self, graph, tumor_state, allowed_lnl_states):
        """Initialize the nodes of the graph."""
        self._nodes: dict[str, Tumor | LymphNodeLevel] = {}

        for node_type, node_name in graph:
            if node_type == "tumor":
                tumor = Tumor(name=node_name, state=tumor_state)
                self._nodes[node_name] = tumor
            elif node_type == "lnl":
                lnl = LymphNodeLevel(name=node_name, allowed_states=allowed_lnl_states)
                self._nodes[node_name] = lnl

        if len(self.tumors) < 1:
            raise ValueError("At least one tumor node must be present in the graph")

        if len(self.lnls) < 1:
            raise ValueError("At least one LNL node must be present in the graph")


    @property
    def nodes(self) -> dict[str, Tumor | LymphNodeLevel]:
        """List of both :py:class:`~Tumor` and :py:class:`~LymphNodeLevel` instances."""
        return self._nodes

    @property
    def tumors(self) -> dict[str, Tumor]:
        """List of all :py:class:`~Tumor` nodes in the graph."""
        return {n: t for n, t in self.nodes.items() if isinstance(t, Tumor)}

    @property
    def lnls(self) -> dict[str, LymphNodeLevel]:
        """List of all :py:class:`~LymphNodeLevel` nodes in the graph."""
        return {n: l for n, l in self.nodes.items() if isinstance(l, LymphNodeLevel)}


    @property
    def allowed_states(self) -> list[int]:
        """Return the list of allowed states for each :py:class:`~LymphNodeLevel`."""
        return next(iter(self.lnls.values())).allowed_states

    @property
    def is_binary(self) -> bool:
        """Indicate if the model is binary.

        Returns ``True`` if all :py:class:`~LymphNodeLevel` instances are binary,
        ``False`` otherwise.
        """
        res = {node.is_binary for node in self.lnls.values()}

        if len(res) != 1:
            raise RuntimeError("Not all lnls have the same number of states")

        return res.pop()

    @property
    def is_trinary(self) -> bool:
        """Returns ``True`` if the graph is trinary, ``False`` otherwise.

        Similar to :py:meth:`~Unilateral.is_binary`."""
        res = {node.is_trinary for node in self.lnls.values()}

        if len(res) != 1:
            raise RuntimeError("Not all lnls have the same number of states")

        return res.pop()


    def _init_edges(
        self,
        graph: dict[tuple[str, str], list[str]],
        on_edge_change: list[callable]
    ) -> None:
        """Initialize the edges of the ``graph``.

        Every one of the provided ``on_edge_change`` list of callback functions is
        called whenever a parameter of an edge is changed. Typically, this is used to
        update the transition tensor of the edge or the transition matrix of the
        :py:class:`lymph.models`.

        When a :py:class:`~LymphNodeLevel` is trinary, it is connected to itself via
        a growth edge.
        """
        self._edges: dict[str, Edge] = {}

        for (_, start_name), end_names in graph.items():
            start = self.nodes[start_name]
            if isinstance(start, LymphNodeLevel) and start.is_trinary:
                growth_edge = Edge(parent=start, child=start, callbacks=on_edge_change)
                self._edges[growth_edge.get_name()] = growth_edge

            for end_name in end_names:
                end = self.nodes[end_name]
                new_edge = Edge(parent=start, child=end, callbacks=on_edge_change)
                self._edges[new_edge.get_name()] = new_edge


    @property
    def edges(self) -> dict[str, Edge]:
        """Iterable of all edges in the graph."""
        return self._edges

    @property
    def tumor_edges(self) -> dict[str, Edge]:
        """List of all tumor :py:class:`~Edge` instances in the graph.

        This contains all edges who's parents are instances of :py:class:`~Tumor` and
        who's children are instances of :py:class:`~LymphNodeLevel`.
        """
        return {n: e for n, e in self.edges.items() if e.is_tumor_spread}

    @property
    def lnl_edges(self) -> dict[str, Edge]:
        """List of all LNL :py:class:`~Edge` instances in the graph.

        This contains all edges who's parents and children are instances of
        :py:class:`~LymphNodeLevel`, including growth edges (if the graph is trinary).
        """
        return {n: e for n, e in self.edges.items() if not e.is_tumor_spread}

    @property
    def growth_edges(self) -> dict[str, Edge]:
        """List of all growth :py:class:`~Edge` instances in the graph.

        Growth edges are only present in trinary models and are arcs where the parent
        and child are the same :py:class:`~LymphNodeLevel` instance. They facilitate
        the change from a micsoscopically positive to a macroscopically positive LNL.
        """
        return {n: e for n, e in self.edges.items() if e.is_growth}


    def __hash__(self) -> int:
        """Return a hash of the graph."""
        hash_res = 0
        for edge in self.edges.values():
            hash_res = hash((hash_res, hash(edge)))

        return hash_res


    def to_dict(self) -> dict[tuple[str, str], set[str]]:
        """Returns graph representing this instance's nodes and egdes as dictionary.

        >>> graph_dict = {
        ...    ('tumor', 'T'): ['II', 'III'],
        ...    ('lnl', 'II'): ['III'],
        ...    ('lnl', 'III'): [],
        ... }
        >>> graph = Representation(graph_dict)
        >>> graph.to_dict() == graph_dict
        True
        """
        res = {}
        for node in self.nodes.values():
            node_type = "tumor" if isinstance(node, Tumor) else "lnl"
            res[(node_type, node.name)] = [o.child.name for o in node.out]
        return res


    def get_mermaid(self) -> str:
        """Prints the graph in mermaid format.

        >>> graph_dict = {
        ...    ('tumor', 'T'): ['II', 'III'],
        ...    ('lnl', 'II'): ['III'],
        ...    ('lnl', 'III'): [],
        ... }
        >>> graph = Representation(graph_dict)
        >>> graph.edges["TtoII"].spread_prob = 0.1
        >>> graph.edges["TtoIII"].spread_prob = 0.2
        >>> graph.edges["IItoIII"].spread_prob = 0.3
        >>> print(graph.get_mermaid())  # doctest: +NORMALIZE_WHITESPACE
        flowchart TD
            T-->|10%| II
            T-->|20%| III
            II-->|30%| III
        <BLANKLINE>
        """
        mermaid_graph = "flowchart TD\n"

        for node in self.nodes.values():
            for edge in node.out:
                mermaid_graph += f"\t{node.name}-->|{edge.spread_prob:.0%}| {edge.child.name}\n"

        return mermaid_graph


    def get_mermaid_url(self) -> str:
        """Returns the URL to the rendered graph."""
        mermaid_graph = self.get_mermaid()
        graphbytes = mermaid_graph.encode("ascii")
        base64_bytes = base64.b64encode(graphbytes)
        base64_string = base64_bytes.decode("ascii")
        url="https://mermaid.ink/img/" + base64_string
        return url


    def get_state(self, as_dict: bool = False) -> dict[str, int] | list[int]:
        """Return the states of the system's LNLs.

        If ``as_dict`` is ``True``, the result is a dictionary with the names of the
        LNLs as keys and their states as values. Otherwise, the result is a list of the
        states of the LNLs in the order they appear in the graph.
        """
        result = {}

        for lnl in self.lnls.values():
            result[lnl.name] = lnl.state

        return result if as_dict else list(result.values())


    def set_state(self, *new_states_args, **new_states_kwargs) -> None:
        """Assign a new state to the system's LNLs.

        The state can either be provided with positional arguments or as keyword
        arguments. In case of positional arguments, the order must be the same as the
        order of the LNLs in the graph. If keyword arguments are used, the keys must be
        the names of the LNLs. The order of the keyword arguments does not matter.

        The keyword arguments override the positional arguments.
        """
        for new_lnl_state, lnl in zip(new_states_args, self.lnls.values()):
            lnl.state = new_lnl_state

        for key, value in new_states_kwargs.items():
            lnl = self.nodes[key]
            if lnl is not None and isinstance(lnl, LymphNodeLevel):
                lnl.state = value


    def _gen_state_list(self):
        """Generates the list of (hidden) states."""
        allowed_states_list = []
        for lnl in self.lnls.values():
            allowed_states_list.append(lnl.allowed_states)

        self._state_list = np.array(list(product(*allowed_states_list)))

    @property
    def state_list(self):
        """Return list of all possible hidden states.

        E.g., for three binary LNLs I, II, III, the first state would be where all LNLs
        are in state 0. The second state would be where LNL III is in state 1 and all
        others are in state 0, etc. The third represents the case where LNL II is in
        state 1 and all others are in state 0, etc. Essentially, it looks like binary
        counting:

        >>> graph = Representation(graph_dict={
        ...     ("tumor", "T"): ["I", "II" , "III"],
        ...     ("lnl", "I"): [],
        ...     ("lnl", "II"): ["I", "III"],
        ...     ("lnl", "III"): [],
        ... })
        >>> graph.state_list
        array([[0, 0, 0],
               [0, 0, 1],
               [0, 1, 0],
               [0, 1, 1],
               [1, 0, 0],
               [1, 0, 1],
               [1, 1, 0],
               [1, 1, 1]])
        """
        try:
            return self._state_list
        except AttributeError:
            self._gen_state_list()
            return self._state_list


    def get_params(
        self,
        as_dict: bool = True,
        as_flat: bool = True,
    ) -> Iterable[float] | dict[str, float]:
        """Return the parameters of the edges in the graph.

        If ``as_dict`` is ``False``, return an iterable of all parameter values. If
        ``as_dict`` is ``True``, return a nested dictionary with the edges' names as
        keys and the edges' parameter dicts as values.

        If ``as_flat`` is ``True``, return a flat dictionary with the T-stages and
        parameters as keys and values, respectively. This is the result of passing the
        nested dictionary to :py:meth:`~lymph.helper.flatten`.
        """
        params = {}
        for edge in self.edges.values():
            params[edge.get_name()] = edge.get_params(as_flat=as_flat)

        if as_flat or not as_dict:
            params = flatten(params)

        return params if as_dict else params.values()


    def set_params(self, *args, **kwargs) -> tuple[float]:
        """Set the parameters of the edges in the graph.

        The arguments are passed to the :py:meth:`~lymph.graph.Edge.set_params` method
        of the edges. Global keyword arguments (e.g. ``"spread"``) are passed to each
        edge's ``set_params`` method. Unused args are returned.

        Specific keyword arguments take precedence over global ones which in turn take
        precedence over positional arguments.

        >>> graph = Representation(graph_dict={
        ...     ("tumor", "T"): ["II" , "III"],
        ...     ("lnl", "II"): ["III"],
        ...     ("lnl", "III"): [],
        ... })
        >>> _ = graph.set_params(0.1, 0.2, 0.3, spread=0.4, TtoII_spread=0.5)
        >>> graph.get_params(as_dict=True)   # doctest: +NORMALIZE_WHITESPACE
        {'TtoII_spread': 0.5,
         'TtoIII_spread': 0.4,
         'IItoIII_spread': 0.4}
        """
        return set_params_for(self.edges, *args, **kwargs)



"""Check functionality of the distribution over diagnose times."""
import unittest
import warnings

import numpy as np
import scipy as sp

from lymph.diagnose_times import Distribution


class FixtureMixin:
    """Mixin that provides fixtures for the tests."""

    @staticmethod
    def binom_pmf(
        support: np.ndarray,
        max_time: int = 10,
        p: float = 0.5,
    ) -> np.ndarray:
        """Binomial probability mass function."""
        if max_time <= 0:
            raise ValueError("max_time must be a positive integer.")
        if len(support) != max_time + 1:
            raise ValueError("support must have length max_time + 1.")
        if not 0. <= p <= 1.:
            raise ValueError("p must be between 0 and 1.")

        return sp.stats.binom.pmf(support, max_time, p)


    def setUp(self):
        self.max_time = 10
        self.array_arg = np.random.uniform(size=self.max_time + 1, low=0., high=10.)
        self.func_arg = lambda support, p=0.5: self.binom_pmf(support, self.max_time, p)


class DistributionTestCase(FixtureMixin, unittest.TestCase):
    """Test the distribution dictionary."""

    def test_frozen_distribution_without_max_time(self):
        """Test the creation of a frozen distribution without providing a max time."""
        dist = Distribution(self.array_arg)
        self.assertFalse(dist.is_updateable)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UserWarning)
            self.assertEqual({}, dist.get_params(as_dict=True))
        self.assertTrue(len(dist.support) == self.max_time + 1)
        self.assertTrue(len(dist.pmf) == self.max_time + 1)
        self.assertTrue(np.allclose(sum(dist.pmf), 1.))

    def test_frozen_distribution_with_max_time(self):
        """Test the creation of a frozen distribution where we provide the max_time."""
        dist = Distribution(self.array_arg)
        self.assertFalse(dist.is_updateable)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UserWarning)
            self.assertEqual({}, dist.get_params(as_dict=True))
        self.assertTrue(len(dist.support) == self.max_time + 1)
        self.assertTrue(len(dist.pmf) == self.max_time + 1)
        self.assertTrue(np.allclose(sum(dist.pmf), 1.))

        self.assertRaises(ValueError, Distribution, self.array_arg, max_time=5)

    def test_updateable_distribution_without_max_time(self):
        """Test the creation of an updateable distribution without providing a max time."""
        self.assertRaises(ValueError, Distribution, self.func_arg)

    def test_updateable_distribution_with_max_time(self):
        """Test the creation of an updateable distribution where we provide the max_time."""
        dist = Distribution(self.func_arg, max_time=self.max_time)
        self.assertTrue(dist.is_updateable)

        dist.set_params(p=0.5)
        self.assertTrue(len(dist.support) == self.max_time + 1)
        self.assertTrue(len(dist.pmf) == self.max_time + 1)
        self.assertTrue(np.allclose(sum(dist.pmf), 1.))

    def test_updateable_distribution_raises_value_error(self):
        """Check that an invalid parameter raises a ValueError."""
        dist = Distribution(self.func_arg, max_time=self.max_time)
        self.assertTrue(dist.is_updateable)
        self.assertRaises(ValueError, dist.set_params, p=1.5)



"""
Test the midline model for the binary case.
"""
import unittest
from typing import Literal

import numpy as np
import pandas as pd

from lymph import models

from . import fixtures


class MidlineSetParamsTestCase(unittest.TestCase):
    """Check that the complex parameter assignment works correctly."""

    def setUp(
        self,
        seed: int = 42,
        graph_size: Literal["small", "medium", "large"] = "small",
        use_mixing: bool = True,
        use_central: bool = True,
        is_symmetric: dict[str, bool] | None = None,
    ) -> None:
        super().setUp()
        self.rng = np.random.default_rng(seed)
        graph_dict = fixtures.get_graph(graph_size)
        if is_symmetric is None:
            is_symmetric = {"tumor_spread": False, "lnl_spread": True}

        self.model = models.Midline(
            graph_dict=graph_dict,
            is_symmetric=is_symmetric,
            use_mixing=use_mixing,
            use_central=use_central,
            use_midext_evo=False,
        )


    def test_init(self) -> None:
        """Check some basic attributes."""
        self.assertTrue(self.model.use_central)
        self.assertTrue(self.model.use_mixing)
        self.assertFalse(self.model.is_trinary)


    def test_set_spread_params(self) -> None:
        """Check that the complex parameter assignment works correctly."""
        params_to_set = {k: self.rng.uniform() for k in self.model.get_params().keys()}
        self.model.set_params(**params_to_set)

        self.assertEqual(
            self.model.central.ipsi.get_tumor_spread_params(),
            self.model.central.contra.get_tumor_spread_params(),
        )
        self.assertEqual(
            self.model.central.ipsi.get_lnl_spread_params(),
            self.model.central.contra.get_lnl_spread_params(),
        )
        self.assertEqual(
            self.model.central.contra.get_lnl_spread_params(),
            self.model.ext.ipsi.get_lnl_spread_params(),
        )
        self.assertEqual(
            self.model.ext.ipsi.get_lnl_spread_params(),
            self.model.noext.ipsi.get_lnl_spread_params(),
        )
        self.assertEqual(
            self.model.ext.ipsi.get_tumor_spread_params(),
            self.model.noext.ipsi.get_tumor_spread_params(),
        )


class MidlineLikelihoodTestCase(unittest.TestCase):
    """Check that the likelihood function works correctly."""

    def setUp(
        self,
        seed: int = 42,
        graph_size: Literal["small", "medium", "large"] = "small",
        use_mixing: bool = True,
        use_central: bool = False,
        use_midext_evo: bool = True,
        is_symmetric: dict[str, bool] | None = None,
    ) -> None:
        super().setUp()
        self.rng = np.random.default_rng(seed)
        graph_dict = fixtures.get_graph(graph_size)
        if is_symmetric is None:
            is_symmetric = {"tumor_spread": False, "lnl_spread": True}

        self.model = models.Midline(
            graph_dict=graph_dict,
            is_symmetric=is_symmetric,
            use_mixing=use_mixing,
            use_central=use_central,
            use_midext_evo=use_midext_evo,
        )
        self.model.set_distribution(
            "early",
            fixtures.create_random_dist(
                type_="frozen",
                max_time=self.model.max_time,
                rng=self.rng,
            ),
        )
        self.model.set_distribution(
            "late",
            fixtures.create_random_dist(
                type_="parametric",
                max_time=self.model.max_time,
                rng=self.rng,
            ),
        )
        self.model.set_modality("pathology", spec=1., sens=1., kind="pathological")
        self.model.load_patient_data(pd.read_csv("./tests/data/2021-clb-oropharynx.csv", header=[0,1,2]))


    def test_likelihood(self) -> None:
        """Check that the likelihood function works correctly."""
        params_to_set = {k: self.rng.uniform() for k in self.model.get_params().keys()}
        self.model.set_params(**params_to_set)

        # Check that the likelihood is a number
        self.assertTrue(np.isscalar(self.model.likelihood()))

        # Check that the likelihood is not NaN
        self.assertFalse(np.isnan(self.model.likelihood()))

        # Check that the log-likelihood is smaller than 0
        self.assertLessEqual(self.model.likelihood(), 0)


class MidlineDrawPatientsTestCase(unittest.TestCase):
    """Check the data generation."""

    def setUp(self) -> None:
        super().setUp()
        self.rng = np.random.default_rng(42)
        graph_dict = {
            ("tumor", "T"): ["A"],
            ("lnl", "A"): ["B"],
            ("lnl", "B"): [],
        }
        self.model = models.Midline(
            graph_dict=graph_dict,
            use_mixing=True,
            use_central=False,
            use_midext_evo=True,
            marginalize_unknown=False,
            unilateral_kwargs={"max_time": 2},
        )
        self.model.set_distribution("early", [0., 1., 0.])
        self.model.set_distribution("late", [0., 0., 1.])
        self.model.set_modality("pathology", spec=1., sens=1., kind="pathological")


    def test_draw_patients(self) -> None:
        """Check that the data generation works correctly."""
        self.model.set_params(
            ipsi_TtoA_spread=1.0,
            contra_TtoA_spread=0.0,
            AtoB_spread=1.0,
            mixing=0.5,
            midext_prob=0.5,
        )
        drawn_data = self.model.draw_patients(
            num=100,
            stage_dist=[0.5, 0.5],
            rng=self.rng,
        )
        self.assertEqual(len(drawn_data), 100)



"""
Full integration test directly taken from the quickstart guide. Aimed at checking the
computed value of the likelihood function.
"""
import unittest

import numpy as np
import scipy as sp

import lymph

from . import fixtures


def late_binomial(support: np.ndarray, p: float = 0.5) -> np.ndarray:
    """Parametrized binomial distribution."""
    return sp.stats.binom.pmf(support, n=support[-1], p=p)


class IntegrationTestCase(fixtures.BinaryUnilateralModelMixin, unittest.TestCase):
    """Run a stripped down version of the quickstart guide."""

    def setUp(self):
        self.graph_dict = fixtures.get_graph(size="medium")
        self.model = lymph.models.Unilateral.binary(graph_dict=self.graph_dict)
        self.model.set_modality("PET", spec=0.86, sens=0.79)
        self.load_patient_data("2021-usz-oropharynx.csv")

        early_fixed = sp.stats.binom.pmf(
            np.arange(self.model.max_time + 1),
            self.model.max_time,
            0.4,
        )
        self.model.set_distribution("early", early_fixed)
        self.model.set_distribution("late", late_binomial)

    def test_likelihood_value(self):
        """Check that the computed likelihood is correct."""
        test_probabilities = [0.02, 0.24, 0.03, 0.2, 0.23, 0.18, 0.18, 0.5]
        llh = self.model.likelihood(given_params=test_probabilities, log=True)
        self.assertAlmostEqual(llh, -586.8723971388224, places=10)
